#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define elementsof(array) ((int)(sizeof(array)/sizeof(array[0])))

static struct map
{
  long upper, lower;
} map[2000];

static int entries;

void unicode_to_utf8(long unicode_ch, char *buf)
{
  char *orig=buf;

  if (unicode_ch<0x80)
  {
    *buf++=(char)unicode_ch;
  }
  else if (unicode_ch<0x7ff)
  {
    *buf++=(char)(0xc0|(unicode_ch>>6));
    *buf++=0x80|(unicode_ch&0x3f);
  }
  else if (unicode_ch<0xffff)
  {
    *buf++=(char)(0xe0|(unicode_ch>>12));
    *buf++=0x80|((unicode_ch>>6)&0x3f);
    *buf++=0x80|(unicode_ch&0x3f);
  }
  else if (unicode_ch<0x10ffff)
  {
    *buf++=(char)(0xf0|(unicode_ch>>18));
    *buf++=0x80|((unicode_ch>>12)&0x3f);
    *buf++=0x80|((unicode_ch>>6)&0x3f);
    *buf++=0x80|(unicode_ch&0x3f);
  }
  else assert(0);
  while (buf<(orig+4)) *buf++='\0';
}

int lower_compare(void const *a_arg, void const *b_arg)
{
  struct map const *a, *b;
  long diff;
  
  a=(struct map const*)a_arg;
  b=(struct map const*)b_arg;
  diff=a->lower-b->lower;
  if (diff<0) diff=-1;
  else if (diff>0) diff=1;
  return (int)diff;
}

int main(int argc, char *argv[])
{
  char ln[256];
  FILE *fp;
  int i;

  fp=fopen("unicode/CaseFolding.txt","r");
  while (fgets(ln, sizeof(ln), fp))
  {
    if ((ln[0]>='0' && ln[0]<='9') || (ln[0]>='A' && ln[0]<='F'))
    {
      char *upper_str, *field_str, *lower_str, *end;
      long upper, lower;

      upper_str=strtok(ln, ";");
      field_str=strtok(NULL, ";");
      lower_str=strtok(NULL, ";");

      upper=strtol(upper_str,&end,16);
      while (*field_str==' ') ++field_str;
      lower=strtol(lower_str,&end,16);

      /* Using F is not technically more complex, but at least for
       * German it converts ß to ss, which is questionable for
       * normalization, but wrong for lower case conversion, as it
       * already is a lower case character. Further it would not
       * allow to toggle case forth and back.
       *
       * The wish to toggle characters also asks for removing
       * pairs with the same lower case letter.
       */
      if (field_str[0]=='C' || field_str[0]=='S')
      {
        int i;

        for (i=0; i<entries && map[i].lower!=lower; ++i);
        if (i==entries)
        {
          assert(entries<elementsof(map));
          map[entries].upper=upper;
          map[entries].lower=lower;
          ++entries;
        }
      }
    }
  }
  fclose(fp);

  fp=fopen("tolower.h","w");
  for (i=0; i<entries; ++i)
  {
    char buf[4];

    unicode_to_utf8(map[i].lower,buf);
    fprintf(fp,"{ 0x%lx, { { (char)0x%x,(char)0x%x,(char)0x%x,(char)0x%x } } },\n",map[i].upper,buf[0]&0xff,buf[1]&0xff,buf[2]&0xff,buf[3]&0xff);
  }
  fclose(fp);

  qsort(map, entries, sizeof(map[0]), lower_compare);

  fp=fopen("toupper.h","w");
  for (i=0; i<entries; ++i)
  {
    char buf[4];

    unicode_to_utf8(map[i].upper,buf);
    fprintf(fp,"{ 0x%lx, { { (char)0x%x,(char)0x%x,(char)0x%x,(char)0x%x } } },\n",map[i].lower,buf[0]&0xff,buf[1]&0xff,buf[2]&0xff,buf[3]&0xff);
  }
  fclose(fp);
}
